/*
 *  uri.c
 *
 *  PostgreSQL Base Type URI (RFC 2396)
 *
 *
 */


#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "utils/builtins.h"
#include "libpq/pqformat.h"
#include "mb/pg_wchar.h"

#include "liburi/uri.h"

/*
 *  Prototypes
 */

PG_FUNCTION_INFO_V1(uri_in);
PG_FUNCTION_INFO_V1(uri_out);
PG_FUNCTION_INFO_V1(uri_recv);
PG_FUNCTION_INFO_V1(uri_send);

PG_FUNCTION_INFO_V1(uri_text);
PG_FUNCTION_INFO_V1(text_uri);

PG_FUNCTION_INFO_V1(uri_uri_table);
PG_FUNCTION_INFO_V1(uri_at_uri);

PG_FUNCTION_INFO_V1(get_uri_scheme);
PG_FUNCTION_INFO_V1(get_uri_user);
PG_FUNCTION_INFO_V1(get_uri_passwd);
PG_FUNCTION_INFO_V1(get_uri_host);
PG_FUNCTION_INFO_V1(get_uri_port);
PG_FUNCTION_INFO_V1(get_uri_path);
PG_FUNCTION_INFO_V1(get_uri_query);
PG_FUNCTION_INFO_V1(get_uri_frag);

PG_FUNCTION_INFO_V1(get_uri_robots);
PG_FUNCTION_INFO_V1(get_uri_root);
PG_FUNCTION_INFO_V1(get_uri_relative);

PG_FUNCTION_INFO_V1(uri_hash);
PG_FUNCTION_INFO_V1(bt_uri_cmp);
PG_FUNCTION_INFO_V1(uri_eq);
PG_FUNCTION_INFO_V1(uri_ne);
PG_FUNCTION_INFO_V1(uri_lt);
PG_FUNCTION_INFO_V1(uri_le);
PG_FUNCTION_INFO_V1(uri_gt);
PG_FUNCTION_INFO_V1(uri_ge);

/*
 *  Commonly used code snippets
 */

static
text *uriString_uriText(char *uriString)
{
    int len = strlen(uriString);
    text *uriText = (text *) palloc(VARHDRSZ + len);
    VARATT_SIZEP(uriText) = VARHDRSZ + len;
    memcpy(VARDATA(uriText),uriString,len);
    return uriText;
}

static
int uri_cmp_internal(text *arg1, text *arg2)
{
    int		result;
    int		len1 = VARSIZE(arg1) - VARHDRSZ,
		len2 = VARSIZE(arg2) - VARHDRSZ;

    result = strncmp(VARDATA(arg1), VARDATA(arg2), len1 < len2 ? len1 : len2);
    if ((result == 0) && (len1 != len2))
	result = (len1 < len2) ? -1 : 1;

    return result;
}

/*
 *  Input / Output functions
 */

Datum
uri_in(PG_FUNCTION_ARGS)
{
    char	*uriString = PG_GETARG_CSTRING(0);
    text	*uriText;
    uri_t	*uriStruct;
    int		len;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(uriString,strlen(uriString));

    if(uriStruct == NULL || uri_consistent(uriStruct) == 0)
	ereport(ERROR,(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),errmsg("invalid input syntax for uri: \"%s\" - %s",uriString,uri_errstr)));

    uriString = uri_uri(uriStruct);
    if(uriStruct->scheme && !strcmp(uriStruct->scheme,"file"))
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_FURI_STYLE);
    else
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_URI_STYLE);
    uriString = uri_unescape(uriString);

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
uri_out(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_uri(uriStruct);
    if(uriStruct->scheme && !strcmp(uriStruct->scheme,"file"))
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_FURI_STYLE);
    else
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_URI_STYLE);

    PG_RETURN_CSTRING(pstrdup(uriString));
}

Datum
uri_recv(PG_FUNCTION_ARGS)
{
        StringInfo      buf = (StringInfo) PG_GETARG_POINTER(0);
        text       *result;
        char       *str;
        int                     nbytes;

        str = pq_getmsgtext(buf, buf->len - buf->cursor, &nbytes);
        result = (text *) palloc(nbytes + VARHDRSZ);
        VARATT_SIZEP(result) = nbytes + VARHDRSZ;
        memcpy(VARDATA(result), str, nbytes);
        pfree(str);
        PG_RETURN_TEXT_P(result);
}

Datum
uri_send(PG_FUNCTION_ARGS)
{
        text       *t = PG_GETARG_TEXT_P(0);
        StringInfoData buf;

        pq_begintypsend(&buf);
        pq_sendtext(&buf, VARDATA(t), VARATT_SIZEP(t) - VARHDRSZ);
        PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

/*
 *  Cast Functions
 */

Datum
uri_text(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString = (char *) DirectFunctionCall1(uri_out,(Datum) uriText);

    PG_FREE_IF_COPY(uriText,0);
    uriText = (text *) DirectFunctionCall1(textin,(Datum) uriString);
    pfree(uriString);

    PG_RETURN_TEXT_P(uriText);
}

Datum
text_uri(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString = (char *) DirectFunctionCall1(textout,(Datum) uriText);

    PG_FREE_IF_COPY(uriText,0);
    uriText = (text *) DirectFunctionCall1(uri_in,(Datum) uriString);
    pfree(uriString);

    PG_RETURN_TEXT_P(uriText);
}

Datum
uri_uri_table(PG_FUNCTION_ARGS)
{
    text		*uriText = PG_GETARG_TEXT_P(0);
    char		*uriString;
    uri_t		*uriStruct;
    char		*result[8];

    TupleDesc           tupdesc;
    AttInMetadata       *attinmeta;
    HeapTuple           tuple;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriString = (char *) DirectFunctionCall1(uri_out,(Datum) uriText);
    PG_FREE_IF_COPY(uriText,0);
    
    uriStruct = uri_object(uriString,strlen(uriString));
    pfree(uriString);

    result[0] = uriStruct->scheme;
    result[1] = uriStruct->user;
    result[2] = uriStruct->passwd;
    result[3] = uriStruct->host;
    result[4] = uriStruct->port;
    result[5] = uriStruct->path;
    result[6] = uriStruct->query;
    result[7] = uriStruct->frag;

    tupdesc = RelationNameGetTupleDesc("pg_catalog.uri_table");
    attinmeta = TupleDescGetAttInMetadata(tupdesc);
    tuple = BuildTupleFromCStrings(attinmeta, result);

    pfree(tupdesc);
    pfree(attinmeta);
    
    PG_RETURN_DATUM(HeapTupleGetDatum(tuple));
}

/*
 *  Hash Operator Class Functions
 */

Datum
uri_hash(PG_FUNCTION_ARGS)
{
    text                *arg;
    int32               result;

    arg = PG_GETARG_TEXT_P(0);
    result = hash_any(VARDATA(arg), VARSIZE(arg) - VARHDRSZ);
    PG_FREE_IF_COPY(arg,0);
    
    return result;
}

/*
 *  Btree Operator Class Functions
 */

Datum
bt_uri_cmp(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = uri_cmp_internal(arg1, arg2);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_INT32(result);
}


Datum
uri_eq(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) == 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

Datum
uri_ne(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) != 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

Datum
uri_lt(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) < 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

Datum
uri_le(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) <= 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

Datum
uri_gt(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) > 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

Datum
uri_ge(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) >= 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

/*
 *  Utility Functions
 */

Datum
uri_at_uri(PG_FUNCTION_ARGS)
{
    text	*uriText1,
		*uriText2;
    uri_t	*uriStruct;
    char	*uriString;

    if(PG_ARGISNULL(0))
	PG_RETURN_NULL();

    if(PG_ARGISNULL(1))
	PG_RETURN_TEXT_P(PG_GETARG_TEXT_P_COPY(0));
    
    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);

    uriText1 = PG_GETARG_TEXT_P(0);
    uriText2 = PG_GETARG_TEXT_P(1);

    uriStruct = uri_object(VARDATA(uriText2),VARATT_SIZEP(uriText2) - VARHDRSZ);
    uriStruct = uri_abs(uriStruct, VARDATA(uriText1), VARATT_SIZEP(uriText1) - VARHDRSZ);

    PG_FREE_IF_COPY(uriText1,0);
    PG_FREE_IF_COPY(uriText2,1);

    uriString = uri_uri(uriStruct);
    if(uriStruct->scheme && !strcmp(uriStruct->scheme,"file"))
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_FURI_STYLE);
    else
	uriString = uri_cannonicalize_string(uriString,strlen(uriString),URI_STRING_URI_STYLE);

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_scheme(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->scheme;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_user(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->user;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_passwd(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->passwd;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_host(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->host;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_port(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->port;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_path(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->path;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_query(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->query;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_frag(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uriStruct->frag;
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_robots(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_robots(uriStruct);
    if(uriString == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_root(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uri_path_set(uriStruct,NULL);
    uri_params_set(uriStruct,NULL);
    uri_query_set(uriStruct,NULL);
    uri_frag_set(uriStruct,NULL);

    if((uri_consistent(uriStruct) == 0) || (uriString = uri_uri(uriStruct)) == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

Datum
get_uri_relative(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    uriStruct = uri_cannonical(uriStruct);
    PG_FREE_IF_COPY(uriText,0);

    uri_scheme_set(uriStruct,NULL);
    uri_user_set(uriStruct,NULL);
    uri_passwd_set(uriStruct,NULL);
    uri_host_set(uriStruct,NULL);
    uri_port_set(uriStruct,NULL);

    if((uri_consistent(uriStruct) == 0) || (uriString = uri_uri(uriStruct)) == NULL)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}
